use File::Find;
use File::Copy;
use File::stat;
use Time::Local;

#------------------------------------------------------------------------------
# Constants
#------------------------------------------------------------------------------
$VERSION='v1.03-pl';
$RELEASEDATE='20.01.2012';
$AUTHOR='Alex Porollo and Vladimir Donskoy';
$M_SPEEDS=10;
$M_PREFIX=10;
$M_FLAGS=255;
$M_IMPLIES=20;
$M_IMPLIES_FLAGS=20;
$M_NETS=255;
$M_PVTS=250;
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Variables
#------------------------------------------------------------------------------
$max_speeds=$max_prefix=$max_exc_prefix=$max_flags=$max_implies=0;
$max_nets=$max_pvts=$owner_zone=$owner_net=$owner_node=$owner_point=0;
$orig_pkt_zone=$orig_pkt_net=$orig_pkt_node=$orig_pkt_point=0;
$dest_pkt_zone=$dest_pkt_net=$dest_pkt_node=$dest_pkt_point=0;
$orig_msg_zone=$orig_msg_net=$orig_msg_node=$orig_msg_point=0;
$def_zone=$def_net=$moving_bad=$moving_ok=$pkt_ok=$confirm_processing=0;
$send_to_owner=$send_to_nc=$send_to_node=$send_to_point=$uppercase=0;
$phonelen=$banner_ok=$lines=$errors=$warnings=$cfgerrors=$notes=0;
$n_zone=$n_net=$n_node=$n_point=$segcr_zone=$segcr_net=$segcr_node=0;
$o_zone=$o_net=$o_node=$o_point=0;
$owner_pkt=$owner_msg=$nc_pkt=$nc_msg=$node_pkt=$node_msg=$point_pkt=0;
$point_msg=$segcr_pkt=$max_n_flags=$enable_mes=$max_dp=$segcr_msg=0;
$owner_sysop_name="Sysop";
$nc_sysop_name="NC";
$from_name="checknl";
$log_level=$scr_level="*";
$path_bad=$path_ok=$path_pkt=$pathname_log='';
$pbanner=$segment_info=$original_string=$movename='';
$n_system=$n_sysop=$n_location=$n_phone=$n_prefix=$n_speed='';
@speeds=@prefix=@exc_prefix=@flags=@implies=@nets=@netphones=@legalpvt=();
@events=@n_flags=%dp=();
@months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Main program
#------------------------------------------------------------------------------
print stderr "Nodelist checker $VERSION, (c) 2005 $AUTHOR\n";
print stderr "Fidonet: 2:5052/5; Internet: ap\@cbank.mari.ru\n";
print stderr "The original C sources: Nodelist checker v1.11, (c) 1994-96 Mikel Lavrentyev\n";
print stderr "Fidonet: 2:5020/35; Internet: mikel\@rinet.ru\n\n";

&load_configuration("checknl.cfg");
&load_configuration("region.cfg");

if (scalar(@ARGV)!=1) {
  print stderr "checknl [filename(s)], for example: checknl net*.*\n";
  exit(0);
}

&write_log('-',"");
&write_log('+',"begin, nodelist checker $VERSION\n");
&write_log('+',"configuration files passed\n");
&write_log('+',"speeds: %d   flags: %d   implies: %d\n",$max_speeds,$max_flags,$max_implies);
&write_log('+',"netphones: %d   legalpvt: %d\n",$max_nets,$max_pvts);
&write_log('+',"command line parameter: %s\n",$ARGV[0]);

@d=(); @d=$ARGV[0];
find(\&FindFiles,@d);
&pkt_end;
&write_log('+',"end\n");
exit(0);
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Subprograms
#------------------------------------------------------------------------------
sub is_flag {
  my $check_flag=shift;
  my $is_flag_ret=$ch=$chp=0;
  for (my $i=0; $i<=$max_flags-1; $i++) {
    $is_flag_ret=$i;
    for (my $j=0; $j<=length($flags[$i]); $j++) {
      $ch=substr($flags[$i],$j,1);
      $chp=substr($check_flag,$j,1);
      last if ($ch eq '*' && $chp);
      next if ($ch eq '?' && $chp);
      if ($ch ne $chp) {
        $is_flag_ret=-1; last;
      }
    }
    if ($is_flag_ret!=-1) {last;}
  }
  return $is_flag_ret;
}

sub load_configuration {
  my $cfgname=shift;
  my $err=$lines=$i=$accept=0;
  open(in,"<$cfgname")||die printf("file not found: %s (configuration)\n",$cfgname);
  while (<in>) {
    $lines++;
    s/\n|\r//g; s/^\s*(.+?)\s*$/$1/;
    if (($_ eq "")|($_=~/^;/)) {next;}
    s/^(.+?)\;.*?$/$1/; s/\t/ /g; s/\s+/ /g;
    @params=split(/ /,$_); my $up=uc($params[0]);
    if ($up eq "SPEED") {
      $i=1;
      while ($params[$i]) {
        if ($max_speeds>$M_SPEEDS-1) {
          printf stderr "file %s, line %3d: max speeds (%d) exceed\n",$cfgname,$lines,$M_SPEEDS;
          $err++; last;
        }
        else {
          $speeds[$max_speeds]=$params[$i];
          $max_speeds++; $i++;
        }
      }
    }
    elsif ($up eq "PREFIX") {
      $i=1;
      while ($params[$i]) {
        if ($max_prefix>$M_PREFIX-1) {
          printf stderr "file %s, line %3d: max prefix (%d) exceed\n",$cfgname,$lines,$M_PREFIX;
          $err++; last;
        }
        else {
          $prefix[$max_prefix]=$params[$i];
          $max_prefix++; $i++;
        }
      }
    }
    elsif ($up eq "DUPES_EXCLUDE") {
      $i=1;
      while ($params[$i]) {
        if ($max_exc_prefix>$M_PREFIX-1) {
          printf stderr "file %s, line %3d: max prefix (%d) exceed\n",$cfgname,$lines,$M_PREFIX;
          $err++; last;
        }
        else {
          $exc_prefix[$max_exc_prefix]=$params[i];
          $max_exc_prefix++; $i++;
        }
      }
    }
    elsif ($up eq "FLAGS") {
      $i=1;
      while ($params[$i]) {
        if ($max_flags>$M_FLAGS-1) {
          printf stderr "file %s, line %3d: max flags (%d) exceed\n",$cfgname,$lines,$M_FLAGS;
          $err++; last;
        }
        else {
          $flags[$max_flags]=$params[$i];
          $max_flags++; $i++;
        }
      }
    }
    elsif ($up eq "IMPLIES") {
      if ($max_implies>$M_IMPLIES-1) {
        printf stderr "file %s, line %3d: max implies (%d) exceed\n",$cfgname,$lines,$M_IMPLIES;
        $err++; next;
      }
      $accept=1;
      if (is_flag($params[1])==-1) {
        $accept=0;
        printf stderr "file %s, line %3d: word \"%s\" not defined as flags before using\n",$cfgname,$lines,$params[1];
        $err++;
      }
      if ($params[2] ne '=') {
        $accept=0;
        printf stderr "file %s, line %3d: illegal \"implies\" define format\n",$cfgname,$lines;
        $err++;
      }
      $i=3;
      while ($params[$i]) {
        if (is_flag($params[$i])==-1) {
          $accept=0;
          printf stderr "file %s, line %3d: word \"%s\" not defined as flags before using in implies\n",$cfgname,$lines,$params[$i];
          $err++; last;
        }
        $i++;
      }
      next if ($accept==0);
      $implies[$max_implies][0]=is_flag($params[1]);
      for ($i=1; $i<=$M_IMPLIES_FLAGS; $i++) {
        $implies[$max_implies][$i]=-1;
      }
      $i=3;
      while ($params[$i]) {
        $implies[$max_implies][$i-2]=is_flag($params[$i]);
        $i++;
      }
      $max_implies++;
    }
    elsif ($up eq "NETWORK") {
      if (scalar(@params)<3) {
        printf stderr "file %s, line %3d: keyword \"networks\" have no parameters, ignored\n",$cfgname,$lines;
        $err++; next;
      }
      $i=2;
      while ($params[$i]) {
        if ($max_nets>$M_NETS-1) {
          printf stderr "file %s, line %3d: max networks (%d) exceed\n",$cfgname,$lines,$M_NETS;
          $err++; last;
        }
        else {
          $_=$params[1];
          ~s/^(\d+)$//; $nets[$max_nets]=$1;
          $netphones[$max_nets]=$params[$i];
          $max_nets++; $i++;
        }
      }
    }
    elsif ($up eq "LEGAL_PVT") {
      $i=1;
      while ($params[$i]) {
        if ($max_pvts>$M_PVTS-1) {
          printf stderr "file %s, line %3d: max legal pvt nodes (%d) exceed\n",$cfgname,$lines,$M_PVTS;
          $err++; last;
        }
        else {
          $_=$params[$i];
          ~s/^(\d+)\/(\d+)$//;
          $legalpvt[$max_pvts][0]=$1; $legalpvt[$max_pvts][1]=$2;
          if (!$legalpvt[$max_pvts][0] || !$legalpvt[$max_pvts][1]) {
            printf stderr "file %s, line %3d: illegal node define \"%s\"\n",$cfgname,$lines,$params[$i];
            $err++; last;
          }
          else {$max_pvts++;}
          $i++ ;
        }
      }
    }
    elsif ($up eq "OWNER_ADDRESS") {
      $_=$params[1];
      ~s/^(\d+):(\d+)\/(\d+)\.?(\d*)$//;
      $owner_zone=$1; $owner_net=$2; $owner_node=$3;
      $owner_point=($4)?$4:0;
      if (!$owner_zone || !$owner_net) {
        printf stderr "file %s, line %3d: illegal address \"%s\"\n",$cfgname,$lines,$params[1];
        $err++; next;
      }
      $owner_sysop_name='';
      for ($i=2; $i<=5; $i++) {
        if (scalar(@params)>$i) {
          $owner_sysop_name.=(($i==2)?'':' ').$params[$i];
        }
        else {last;}
      }
    }
    elsif ($up eq "ORIG_PKT") {
      $_=$params[1];
      ~s/^(\d+):(\d+)\/(\d+)\.?(\d*)$//;
      $orig_pkt_zone=$1; $orig_pkt_net=$2; $orig_pkt_node=$3;
      $orig_pkt_point=($4)?$4:0;
      if (!$orig_pkt_zone || !$orig_pkt_net) {
        printf stderr "file %s, line %3d: illegal address \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "DEST_PKT") {
      $_=$params[1];
      ~s/^(\d+):(\d+)\/(\d+)\.?(\d*)$//;
      $dest_pkt_zone=$1; $dest_pkt_net=$2; $dest_pkt_node=$3;
      $dest_pkt_point=($4)?$4:0;
      if (!$dest_pkt_zone || !$dest_pkt_net) {
        printf stderr "file %s, line %3d: illegal address \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "ORIG_MSG") {
      $_=$params[1];
      ~s/^(\d+):(\d+)\/(\d+)\.?(\d*)$//;
      $orig_msg_zone=$1; $orig_msg_net=$2; $orig_msg_node=$3;
      $orig_msg_point=($4)?$4:0;
      if (!$orig_msg_zone || !$orig_msg_net) {
        printf stderr "file %s, line %3d: illegal address \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
      $from_name='';
      for ($i=2; $i<=5; $i++) {
        if (scalar(@params)>$i) {
          $from_name.=(($i==2)?'':' ').$params[$i];
        }
        else {last;}
      }
    }
    elsif ($up eq "DEF_ZONE_NET") {
      $_=$params[1];
      ~s/^(\d+):(\d+)$//;
      $def_zone=$1; $def_net=$2;
      if (!$def_zone || !$def_net) {
        printf stderr "file %s, line %3d: illegal address \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "PATH_BAD_SEGMENTS") {
      $path_bad=$params[1];
      if ($path_bad!~/\/$/) {$path_bad.='/';}
      $moving_bad=1;
    }
    elsif ($up eq "PATH_OK_SEGMENTS") {
      $path_ok=$params[1];
      if ($path_ok!~/\/$/) {$path_ok.='/';}
      $moving_ok=1;
    }
    elsif ($up eq "PATH_FOR_PACKETS") {
      $path_pkt=$params[1];
      if ($path_pkt!~/\/$/) {$path_pkt.='/';}
      $pkt_ok=1;
    }
    elsif ($up eq "PATHNAME_LOG") {
      $pathname_log=$params[1];
    }
    elsif ($up eq "LOG_LEVEL") {
      $log_level=$params[1];
    }
    elsif ($up eq "SCR_LEVEL") {
      $scr_level=$params[1];
    }
    elsif ($up eq "SEND_TO_OWNER") {
      if (uc($params[1]) eq "YES") {$send_to_owner=1;}
      elsif (uc($params[1]) eq "NO") {$send_to_owner=0;}
      else {
        printf stderr "file %s, line %3d: unknown word \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "SEND_TO_NC") {
      if (uc($params[1]) eq "YES") {$send_to_nc=1;}
      elsif (uc($params[1]) eq "NO") {$send_to_nc=0;}
      else {
        printf stderr "file %s, line %3d: unknown word \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "SEND_TO_NODE") {
      if (uc($params[1]) eq "YES") {$send_to_node=1;}
      elsif (uc($params[1]) eq "NO") {$send_to_node=0;}
      else {
        printf stderr "file %s, line %3d: unknown word \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "SEND_TO_POINT") {
      if (uc($params[1]) eq "YES") {$send_to_point=1;}
      elsif (uc($params[1]) eq "NO") {$send_to_point=0;}
      else {
        printf stderr "file %s, line %3d: unknown word \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "CONFIRM_PROCESSING") {
      if (uc($params[1]) eq "YES") {$confirm_processing=1;}
      elsif (uc($params[1]) eq "NO") {$confirm_processing=0;}
      else {
        printf stderr "file %s, line %3d: unknown word \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "UPPERCASE") {
      if (uc($params[1]) eq "YES") {$uppercase=1;}
      elsif (uc($params[1]) eq "NO") {$uppercase=0;}
      else {
        printf stderr "file %s, line %3d: unknown word \"%s\"\n",$cfgname,$lines,$params[1];
        $err++;
      }
    }
    elsif ($up eq "EVENT") {
      $_=$params[1];
      ~s/^(\d+)$//; $i=$1;
      $events[$i]=uc(substr($params[2],0,1)) if ($i=~/[0-9]/);
    }
    elsif ($up eq "PHONELEN") {
      $_=$params[1];
      ~s/^(\d+)$//; $phonelen=$1;
    }
    elsif ($up eq "BANNER") {
      if (!open(ban,"<$params[1]")) {
        printf stderr "cannot open banner file %s for read\n",$params[1];
        $err++;
      }
      else {
        $banner_ok=1;
        while (<ban>) {$pbanner.=$_;}
        close(ban);
      }
    }
    else {
      printf stderr "file %s, line %3d: illegal keyword \"%s\"\n",$cfgname,$lines,$params[0];
      $err++;
    }
  }
  close(in);
  if ($err>0) {
    printf stderr "\n%d error%s encountered, aborted\n",$err,($err==1)?'':'s';
    exit(2);
  }
  printf stderr "configuration file %s passed\n",$cfgname;
}

sub get_msgid {
  my $timer=timelocal(localtime);
  if ($timer!=int($timer/2)*2) {$timer++;} # odd number (for seconds)
  my $msgid=0;
  my $msgid_file="MSGID.DAT";
  if (open(file,"<$msgid_file")) {
    $msgid=hex((split(/ /,<file>))[0])+2;
    close(file);
  }
  if ($msgid<$timer) {$msgid=$timer;}
  open(file,">$msgid_file");
  printf file "%0X ; this is an last number of msgid used by checknl utility\n",$msgid;
  close(file);
  return sprintf("%0X",$msgid);
}

sub write_log {
  my $prefix=shift;
  my @format=();
  foreach my $k (@_) {push(@format,$k);}

  my $log_ok=$scr_ok=0;
  if (($log_level eq '*')||($prefix=~/[$log_level]/)) {$log_ok=1};
  if (($scr_level eq '*')||($prefix=~/[$scr_level]/)) {$scr_ok=1};
  return if (!$log_ok && !$scr_ok );

  if (!open(out,">>$pathname_log")) {return;}

  if ($prefix eq '-') {
    print out "\n";
    close(out);
    return;
  }

  ($sec,$min,$hour,$day,$month)=(localtime)[0..4];
  if ($log_ok) {
    printf out "%s %02d/%02d %02d:%02d:%02d CHNL ",$prefix,$day,$month+1,$hour,$min,$sec;
    printf out @format;
  }
  if ($scr_ok) {
    printf "%s ",$prefix;
    printf @format;
  }
  close(out);
}

sub errors_out {
  my $event=shift;
  my $str=shift;
  my $str1;
  if ($events[$event] eq 'I') {return;}
  if ($events[$event] eq 'B') {
    &write_log('@',"break processing: %s\n",$str);
    $errors++;
    return;
  }
  &write_log('#',"(%s) %d:%d/%d.%d - %s\n",$events[$event],$n_zone,$n_net,$n_node,$n_point,$str);
  if (($n_zone!=$o_zone) || ($n_net!=$o_net) || ($n_node!=$o_node) || ($n_point!=$o_point)) {
    $o_zone=$n_zone; $o_net=$n_net; $o_node=$n_node; $o_point=$n_point;
    &msg_begin($n_zone,$n_net,$n_node,$n_point);
    &msg_out(sprintf("\r%s\r",$segment_info));
    &msg_out(sprintf("creator: node %d:%d/%d\r\r",$segcr_zone,$segcr_net,$segcr_node));
    $str1=sprintf("original %s line:\r%s\r\rparsed %s line:\r",($n_point)?"pointlist":"nodelist",$original_string,($n_point)?"pointlist":"nodelist");
    $str1.=(($n_point==0)?sprintf("node %d:%d/%d, ",$n_zone,$n_net,$n_node):sprintf("point %d:%d/%d.%d, ",$n_zone,$n_net,$n_node,$n_point)).'system "'.$n_system.'", sysop "'.$n_sysop;
    $str1.="\"\rlocation \"".$n_location.'", phone "'.$n_phone;
    $str1.="\"\rprefix \"".$n_prefix.'", speed "'.$n_speed;
    $str1.='", flags "';
    my $i=0;
    while ($i<$max_n_flags) {
      $str1.=$n_flags[$i];
      if ($i<$max_n_flags-1) {$str1.=',';}
      $i++;
    }
    $str1.='"'."\r\r";
    &msg_out($str1);
  }
  if ($events[$event] eq 'E') {
    $errors++;
    $str1="error: ";
  }
  elsif ($events[$event] eq 'W') {
    $warnings++;
    $str1="warning: ";
  }
  elsif ($events[$event] eq 'C') {
    $cfgerrors++;
    $str1="checknl configuration file error: ";
  }
  elsif ($events[$event] eq 'N') {
    $notes++;
    $str1="note: ";
  }
  else {
    $str1=sprintf("internal error, unknown event %s: ",$events[$event]);
  }
  $str1.="$str\r";
  &msg_out($str1);
}

sub pkt_create {
  my $msgid=get_msgid();
  my $pktname=sprintf("%s%s.pkt",$path_pkt,$msgid);
  &write_log('{',"creating packet %s%s.pkt\n",$path_pkt,$msgid);

  if (!open($msgid,">$pktname")) {
    &write_log('!',"fatal error: cannot create packet %s for write\n",$pktname);
    exit(0);
  }

  ($sec,$min,$hour,$day,$month,$year)=(localtime)[0..5];

  my $typedef='s s s s s s s s s A2 s s A1 A1 A8 s s s A1 A1 A1 A1 A1 A1 s s s s A4';
  my $pktf=pack($typedef,$orig_pkt_node,$dest_pkt_node,
             $year-100,$months[$month],$day,$hour,$min,$sec,0,"\x02\x00",
             ($dest_pkt_point)?-1:$orig_pkt_net,$dest_pkt_net,
             "\x00","\x00","\x00\x00\x00\x00\x00\x00\x00\x00",
             $orig_pkt_zone,$dest_pkt_zone,
             ($dest_pkt_point==0)?0:$orig_pkt_net,
             "\x00","\x01","\x00","\x00","\x01","\x00",
             $orig_pkt_zone,$dest_pkt_zone,$orig_pkt_point,$dest_pkt_point,
             "\x00\x00\x00\x00");
  binmode($msgid);
  print $msgid $pktf;
  return $msgid;
}

sub pkt_end {
  &write_log('{',"closing all packets\n");
  $owner_pkt=pkt_close($owner_pkt);
  $nc_pkt=pkt_close($nc_pkt);
  $node_pkt=pkt_close($node_pkt);
  $point_pkt=pkt_close($point_pkt);
  $segcr_pkt=pkt_close($segcr_pkt);
}

sub pkt_close {
  my $fileout=shift;
  if ($fileout ne "0") {
    print $fileout "\x00\x00";
    close($fileout);
    return 0;
  }
}

sub msg_begin {
  my $zone=shift;
  my $net=shift;
  my $node=shift;
  my $point=shift;

  if (!$enable_mes) {return;}

  $owner_msg=$nc_msg=$node_msg=$point_msg=0;

  ($owner_pkt,$owner_msg)=msg_create(
      $owner_zone,$owner_net,$owner_node,$owner_point,
      $owner_pkt,$owner_msg,
      "about errorneous node",$send_to_owner,$owner_sysop_name) ;
  if (!$point) {
    ($nc_pkt,$nc_msg)=msg_create(
      $zone,$net,0,0,
      $nc_pkt,$nc_msg,
      "about errorneous node into your nodelist segment",
      $send_to_nc,$nc_sysop_name);
#    ($node_pkt,$node_msg)=msg_create(
#      $zone,$net,$node,0,
#      $node_pkt,$node_msg,
#      "about your errorneous node into nodelist segment",
#      $send_to_node,$n_sysop);
  }
  else {
    ($node_pkt,$node_msg)=msg_create(
      $zone,$net,$node,0,
      $node_pkt,$node_msg,
      "about errorneous point into your pointlist segment",
      $send_to_node,"Sysop");
    ($point_pkt,$point_msg)=msg_create(
      $zone,$net,$node,$point,
      $point_pkt,$point_msg,
      "about your errorneous point into pointlist segment",
      $send_to_point,$n_sysop);
  }
}

sub msg_create {
  my $zone=shift;
  my $net=shift;
  my $node=shift;
  my $point=shift;
  my $fileout=shift;
  my $msg=shift;
  my $subject=shift;
  my $send_to=shift;
  my $dest_sysop_name=shift;

  if (!$send_to && !$path_pkt) {return;}

  if ($fileout eq '0') {$fileout=pkt_create();}
  my $msg=($point)?2:1;
  my $msgid=get_msgid();
  my $msgf='';

  &write_log('[',"creating message for %d:%d/%d.%d, msgid %s\n",$zone,$net,$node,$point,$msgid);

  ($sec,$min,$hour,$day,$month,$year)=(localtime)[0..5];

  my $typedef='A2 s s s s s s A20';
  $msgf=pack($typedef,"\x02\x00",$orig_msg_node,$node,$orig_msg_net,$net,0,0,
     sprintf("%02d %03s %02d  %02d:%02d:%02d",$day,$months[$month],$year-100,$hour,$min,$sec));
  print $fileout "$msgf$dest_sysop_name\x00$from_name\x00$subject\x00";
  printf $fileout "\x01MSGID: %d:%d/%d.%d %8s\r",$orig_msg_zone,$orig_msg_net,$orig_msg_node,$orig_msg_point,$msgid;
  printf $fileout "\x01PID: CheckNL $VERSION\r";
  if ($orig_msg_point) {printf $fileout "\01xFMPT %d\r",$orig_msg_point;}
  if ($point) {printf $fileout "\01xTOPT %d\r",$point;}
  printf $fileout "\01xINTL %d:%d/%d %d:%d/%d\r",$zone,$net,$node,$orig_msg_zone,$orig_msg_net,$orig_msg_node;
  return ($fileout,$msg);
}

sub msg_out {
  my $string=shift;
  &msg_update($owner_pkt,$owner_msg,$string);
  &msg_update($nc_pkt,$nc_msg,$string);
  &msg_update($node_pkt,$node_msg,$string);
  &msg_update($point_pkt,$point_msg,$string);
}

sub msg_update {
  my $fileout=shift;
  my $msg=shift;
  my @format=();
  foreach my $k (@_) {push(@format,$k);}
  if (($fileout ne "0") & ($msg!=0)) {printf $fileout @format;}
}

sub msg_end {
  $owner_msg=msg_close($owner_pkt,$owner_msg);
  $nc_msg=msg_close($nc_pkt,$nc_msg);
  $node_msg=msg_close($node_pkt,$node_msg);
  $point_msg=msg_close($point_pkt,$point_msg);
}

sub msg_close {
  my $fileout=shift;
  my $msg=shift;
  if (($fileout eq "0") || ($msg==0)) {return;}
  if ($banner_ok) {
    &msg_update($fileout,$msg,"\r");
    &msg_update($fileout,$msg,(""x79)."\r");
    &msg_update($fileout,$msg,$pbanner);
    &msg_update($fileout,$msg,(""x79)."\r");
  }
  &msg_update($fileout,$msg,"--- nodelist checker $VERSION\r");
  &msg_update($fileout,$msg," * Origin: released at $RELEASEDATE by $AUTHOR (%d:%d/%d.%d)\r",$orig_msg_zone,$orig_msg_net,$orig_msg_node,$orig_msg_point);
  print $fileout "\x00";
  return 0;
}

sub processed {
  printf stderr "* lines: %6u   errors: %5u   warnings: %5u   cfgerrors: %2u   notes: %4u\n",$lines,$errors,$warnings,$cfgerrors,$notes;
}

sub check_text_field  {
  my $f=shift;
  my $name=shift;
  if (length($f)==0) {
    &errors_out(15,sprintf("%s field is empty",$name));
    return;
  }
  if ($f=~/^\s/) {
    &errors_out(16,sprintf("%s field contain leading space%s",$name,($f=~/^\s\s/)?"s":""));
  }
  if ($f=~/\s$/) {
    &errors_out(16,sprintf("%s field contain trailing space%s",$name,($f=~/\s\s$/)?"s":""));
  }
  if ($f!~/^[\x20-\x7e]+$/) {
    &errors_out(17,sprintf("%s field contain not valid character",$name));
  }
}

sub strip_phone {
  my $phone=shift;
  my $retflag=0;
  if ($phone!~/^[0-9-]+$/) {$retflag=1;}
  $phone=~s/[^0-9]//g;
  return ($retflag,$phone);
}

sub check_nodelist {
  my $nlname=shift;
  my $sname=$nlname;
  $sname=~s/.*\/(.+?)$/$1/;

  my $old_disabled_net=$header_ok=$i=$q=$ok=$net_ok=$net_pointer=0;
  my $uflag=$retflag=0;
  my $string=$n_phone_s=$netphones_s='';
  my @exist_n_flags=();
  $lines=$errors=$warnings=$cfgerrors=$notes=$n_node=0;
  $segcr_zone=$segcr_net=$segcr_node=$is_eon=$is_cm=0;
  $n_zone=$def_zone;
  $n_net=$def_net;

  &write_log('@',"processing segment \"%s\"\n",$sname);

  if (!open(in,"<$nlname")) {
     &write_log('!',"cannot open \"%s\" for read\n",$nlname);
     return;
  }

  $sf=stat($nlname);
  ($fdss,$fdmi,$fdhh,$fddd,$fdmm,$fdyy)=(localtime $sf->mtime())[0..5];
  $fdmm++; $fdyy+=1900;
  $segment_info.=sprintf(", %ld bytes, dated %02d-%02d-%04d %02d:%02d:%02d",$sf->size(),$fddd,$fdmm,$fdyy,$fdhh,$fdmi,$fdss);
  &write_log('@',"%s\n",$segment_info);

  while (<in>) {
    s/\n|\r//g; s/^\s*(.+?)\s*$/$1/;
    next if ($_=~/^\x1A/);
    $lines++;
    next if ($_=~/^;/);
    &msg_end;
    if (int($lines/100)*100==$lines) {&processed;}
    $original_string=$_;
    ~s/_/ /g;
    @nparam=split(/,/,$_);
    if (scalar(@nparam)==0) {next;}  # this is a comment string
    if (scalar(@nparam)<2) {
      &errors_out(14,"not a nodelist segment! (invalid string)");
      break ;
    }
    if ($nparam[0]!~/^\d+$/) {
      $n_prefix=$nparam[0];
      for ($i=0; $i<=scalar(@nparam)-1; $i++) {
        $nparam[$i]=$nparam[$i+1];
      }
    }
    $n_prefix="[none]" if (!$n_prefix);
    if ($n_prefix eq "Point") {
      $n_point=$nparam[0];
      $n_point=~s/^(\d+)$/$1/;
    }
    else {
      $n_node=$nparam[0];
      $n_node=~s/^(\d+)$/$1/;
      $n_point=0;
    }
    $n_system=$nparam[1];
    $n_location=$nparam[2];
    $n_sysop=$nparam[3];
    $n_phone=$nparam[4];
    $n_speed=$nparam[5];

    if ($n_prefix eq "Host") {
      $n_net=$n_node;
      $n_node=0;
      $header_ok=1;
    }
    elsif ($n_prefix eq "Region") {
      $n_net=$n_node;
      $n_node=0;
      $header_ok=1;
    }
    elsif ($n_prefix eq "Zone") {
      $n_zone=$n_node;
      $n_net=$n_node;
      $n_node=0;
      $header_ok=1;
    }
    elsif ($n_prefix eq "Boss") {
      $_=$nparam[1];
      ~s/^(\d+):(\d+)\/(\d+)$//;
      $n_zone=$1; $n_net=$2; $n_node=$3;
      $header_ok=1;
      next;
    }

    # test of header
    if (!$header_ok) {
      &errors_out(14,"not a nodelist segment! (valid header not present)");
      last;
    }
    if (!$n_node) {$nc_sysop_name=$n_sysop;}

    # checking for enable message writing
    $i=0; $enable_mes=0; $net_pointer=0;
    while ($i<$max_nets) {
      if ($n_net==$nets[$i]) {$enable_mes=1;}
      $net_pointer=$i;
      $i++;
    }
    if (!$enable_mes && $old_disabled_net!=$n_net) {
      $old_disabled_net=$n_net;
      &errors_out(8,sprintf("net %d is not defined, disable messages for this net",$n_net));
    }

    # copying segment creator address for future use
    if (!$segcr_zone && $enable_mes) {
      $segcr_zone=$n_zone;
      $segcr_net=$n_net;
      $segcr_node=$n_node;
    }

    # parsing n_flags
    $uflag=$max_n_flags=0; $i=6;
    while ($i<scalar(@nparam)-1) {
      if (!$uflag) {
        if ($nparam[$i]=~/^U/) {$uflag=1;}
        if ($nparam[$i] eq "U") {
          $i++; next;
        }
        $string=$nparam[$i];
      }
      else {$string="U$nparam[$i]";}
      if ($uppercase) {$string=uc($string)};
      $n_flags[$max_n_flags]=$string;
      $max_n_flags++; $i++;
    }

    # check for alphanumeric characters
    if ($original_string!~/^[\x0a\x21-\x7e]+$/) {
      &errors_out(17,sprintf("original string contain not a valid character"));
    }

    # check for comma at end of the string
    if ($original_string=~/\,$/) {
      &errors_out(17,sprintf("comma on the end of original string"));
    }

    # illegal Pvt nodes
    if ($n_prefix eq "Pvt") {
      $ok=$i=0;
      while ($i<$max_pvts) {
        if (($legalpvt[$i][0]==$n_net) && ($legalpvt[$i][1]==$n_node)) {
          $ok=1; last;
        }
        $i++;
      }
      &errors_out(5,"illegal Pvt node") if (!$ok);
    }

    # unknown prefix
    $ok=$i=0;
    while ($i<$max_prefix) {
      if ($n_prefix eq $prefix[$i]) {$ok=1; last;}
      $i++;
    }
    &errors_out(7,sprintf("unknown prefix \"%s\"",$n_prefix)) if (!$ok);

    # unknown speed
    $ok=$i=0;
    while ($i<$max_speeds) {
      if ($n_speed eq $speeds[$i]) {$ok=1; last;}
      $i++;
    }
    &errors_out(1,sprintf("unknown speed \"%s\"",$n_speed)) if (!$ok);

    # unknown/double flags
    for ($i=0; $i<$M_FLAGS; $i++) {$exist_n_flags[$i]=0;}
    $i=$is_eon=$is_cm=0;
    while ($i<$max_n_flags) {
      $q=is_flag($n_flags[$i]);
      if ($q==-1) {
        &errors_out(2,sprintf("unknown flag \"%s\"",$n_flags[$i]));
        $i++; next;
      }
      if ($exist_n_flags[$q]==1) {
        &errors_out(3,sprintf("double flag \"%s\"",$n_flags[$i]));
        $i++; next;
      }
      $is_eon=1 if ($n_flags[$i]=~/IEM|ITX|IUC|IMI|ISE/);
      $is_cm=1 if ($n_flags[$i]=~/ICM|CM/);
      $exist_n_flags[$q]=1; $i++;
    }
    if ($is_eon && !$is_cm) {
      &errors_out(12,"Non-CM EON");
      next;
    }

    # flags implies
    $i=0;
    while ($i<$max_implies) {
      if ($exist_n_flags[$implies[$i][0]]==0) {  # no main implies flag
        $i++; next;
      }
      $ok=0; $string=""; $j=1;
      while ($implies[$i][$j]!=-1) {
        if ($exist_n_flags[$implies[$i][$j]]==1) {  # redundant flag
          $string.='"'.$flags[$implies[$i][$j]].'", ';
          $ok++;
        }
        $j++;
      }
      if ($ok) {
         $string="redundant flag".(($ok>1)?'s':'')." $string";
         $string.=' implies into flag "'.$flags[$implies[$i][$0]].'"';
         &errors_out(4,$string);
      }
      $i++;
    }

    # check for alphanumeric characters in text fields
    &check_text_field($n_system,"system");
    &check_text_field($n_location,"location");
    &check_text_field($n_sysop,"sysop" );

    # phones checking
    ($retflag,$n_phone_s)=strip_phone($n_phone);
    if ($n_phone=~/^00\-/) {
      if ($n_phone eq "00-00-000000") {
#        if ($n_prefix eq "Host") {
#          &errors_out(12,"using \"00-00-000000\" with prefix \"Host\"");
#        }
      }
      else {
	&errors_out(12,"illegal phone for IP-only");
      }
      next;
    }
    elsif ($n_phone eq "-Unpublished-") {
      if (!(($n_prefix eq "Pvt") || ($n_prefix eq "Point"))) {
        &errors_out(12,"using \"-Unpublished-\" without prefix \"Pvt\" or \"Point\"");
      }
      next;
    }
    elsif (length($n_phone_s)==0) {
      &errors_out(11,sprintf("using word \"%s\" instead of \"-Unpublished-\"",$n_phone));
      next;
    }
    elsif ($retflag) {
      &errors_out(13,sprintf("illegal symbols into phone \"%s\"",$n_phone));
    }
    if (!$retflag && ($phonelen!=0) && (length($n_phone_s)!=$phonelen)) {
      &errors_out(10,sprintf("length of phone \"%s\" is not %d",$n_phone,$phonelen));
    }

    $ok=$net_ok=$i=0;
    while ($i<$max_nets) {
      if ($nets[$i]==$n_net) {
        $net_ok=1;
        ($retflag,$netphones_s)=strip_phone($netphones[$i]);
        if ($n_phone_s=~/^$netphones_s/) {
          $ok=1; last;
        }
      }
      $i++;
    }
    if (!$ok) {
      if ($net_ok) {
         if (!$n_point) {&errors_out(6,sprintf("illegal phone \"%s\", outside from net %u",$n_phone,$n_net));}
         $string=sprintf("net %u phone(s) area(s):",$n_net);
         $i=0;
         while($i<$max_nets) {
           if ($n_net==$nets[$i]) {
             $string.=" $netphones[$i]";
           }
           $i++;
         }
         if (!$n_point) {&errors_out(0,$string);}
      }
      elsif (($n_node==0)&(!$n_point)) {&errors_out(8,sprintf("phone area for net %u is not defined",$n_net));}
      # back-search
      $i=1;
      while ($i<$max_nets) {
        ($retflag,$netphones_s)=strip_phone($netphones[$i]);
        if (($n_phone_s=~/^$netphones_s/)&(!$n_point)) {
          &errors_out(0,sprintf("this phone included into phone area \"%s\" for net %u",$netphones[$i],$nets[i])); # for information!
        }
        $i++;
      }
    }

    # check phone/address duplicates
    if ($n_node==0) {$max_dp=0;}
    else {
      $ok=1; $i=0;
      $dp{$max_dp}{'is_hub'}=0;
      while ($i<$max_exc_prefix) {
        if ($n_prefix eq $exc_prefix[$i]) {
          if ($n_prefix eq "Hub") {
            $dp{$max_dp}{'is_hub'}=1; last;
          }
          else {
            $ok=0; last;
          }
        }
        $i++;
      }
      if ($ok) {
        $i=0;
        while ($i<$max_dp) {
          if (($n_phone_s eq $dp{$i}{'phone'}) && ((!$dp{$max_dp}{'is_hub'} && !$dp{$i}{'is_hub'} || ($dp{$max_dp}{'is_hub'} && $dp{$i}{'is_hub'})) & (!$n_point))) {
            &errors_out(9,sprintf("duplicate %s phone \"%s\" with %s %u/%u",($dp{$max_dp}{'is_hub'})?"hub":"node",$n_phone,($dp{$i}{'is_hub'})?"hub":"node",$dp{$i}{'net'},$dp{$i}{'node'}));
          }
          elsif (($dp{$max_dp}{'net'} eq $dp{$i}{'net'}) && ($dp{$max_dp}{'node'} eq $dp{$i}{'node'}) && ($dp{$max_dp}{'point'} eq $dp{$i}{'point'})) {
            &errors_out(18,sprintf("duplicate address %d/%d.%d",$dp{$max_dp}{'net'},$dp{$max_dp}{'node'},$dp{$max_dp}{'point'}));
          }
          $i++;
        }
        $dp{$max_dp}{'phone'}=$n_phone_s;
        $dp{$max_dp}{'net'}=$n_net;
        $dp{$max_dp}{'node'}=$n_node;
        $dp{$max_dp}{'point'}=$n_point;
        $max_dp++;
        if ($max_dp>=$MAX_DP) {$max_dp--};
      }
    }
  }
  &msg_end;
  close(in);

  &write_log('&',"%s - lines     : %5u\n",$sname,$lines);
  &write_log('&',"%s - errors    : %5u\n",$sname,$errors);
  &write_log('&',"%s - warnings  : %5u\n",$sname,$warnings);
  &write_log('&',"%s - cfgerrors : %5u\n",$sname,$cfgerrors);
  &write_log('&',"%s - notes     : %5u\n",$sname,$notes);

  &processed;
}

sub FindFiles {
  if (-d) {return;}
  my $fname=$sname=$File::Find::name;

  $sname=$_;
  $segment_info=sprintf("segment: %s",$sname);
  &check_nodelist($fname);
  if ((!$errors && $moving_ok) || ($errors && $moving_bad)) {
    $movename=sprintf("%s%s",($errors)?$path_bad:$path_ok,$sname);
    &write_log(($errors)?'>':'<',"moving %s segment: %s to %s\n",($errors)?"bad":"ok",$fname,$movename);
    if (-e $movename) {unlink($movename);}
    copy($fname,$movename);
    if (-e $movename) {
      utime(time,stat($fname)->mtime,$movename);
      unlink($fname);
    }
    else {
      &write_log('!',"cannot move file, OS error\n");
    }
  }
  else {
    &write_log('|',"moving %s not defined\n",($errors)?"bad":"ok");
  }

  if ($confirm_processing && $segcr_zone) {
    ($segcr_pkt,$segcr_msg)=&msg_create(
        $segcr_zone,$segcr_net,$segcr_node,0,
        $segcr_pkt,$segcr_msg,
        $segment_info,$confirm_processing,
        ($segcr_node)?"Sysop":$nc_sysop_name);
    &msg_update($segcr_pkt,$segcr_msg,"\r%s\r\r",$segment_info);
    &msg_update($segcr_pkt,$segcr_msg,"your %s segment contains:\r\r",($segcr_node)?"pointlist":"nodelist");
    &msg_update($segcr_pkt,$segcr_msg,"lines    : %6u\r",$lines);
    &msg_update($segcr_pkt,$segcr_msg,"errors   : %6u %s\r",$errors,($errors)?"":"(none)");
    &msg_update($segcr_pkt,$segcr_msg,"warnings : %6u %s\r",$warnings,($warnings)?"":"(none)");
    &msg_update($segcr_pkt,$segcr_msg,"notes    : %6u %s\r",$notes,($notes)?"":"(none)");
    &msg_update($segcr_pkt,$segcr_msg,"\r");
    &msg_update($segcr_pkt,$segcr_msg,"result of checking: segment is %s\r",($errors)?"bad":"ok");
    &msg_close($segcr_pkt,$segcr_msg);
  }
}
